@@ -65,7 +65,9 @@ following parameters:
 
 .. rubric:: Example
 
->>> tree example/sklearn_iris/mlruns/run1/outputs/linear-lr
+::
+
+    tree example/sklearn_iris/mlruns/run1/outputs/linear-lr
 
 ::
 
@@ -77,7 +79,9 @@ following parameters:
   │   └── model.pkl
   └── mlflow_env.yml
 
->>> cat example/sklearn_iris/mlruns/run1/outputs/linear-lr/MLmodel
+::
+
+    cat example/sklearn_iris/mlruns/run1/outputs/linear-lr/MLmodel
 
 ::
 
@@ -189,24 +193,30 @@ You may prefer the second, lower-level workflow for the following reasons:
 """
 
 import importlib
-import logging
+
 import numpy as np
 import os
 import pandas
-import shutil
+import yaml
 from copy import deepcopy
+import logging
 
+from typing import Any, Union
 import mlflow
 import mlflow.pyfunc.model
 import mlflow.pyfunc.utils
-from mlflow.models import Model
+from mlflow.models import Model, ModelSignature, ModelInputExample
+
+from mlflow.models.utils import _save_example
 from mlflow.pyfunc.model import PythonModel, PythonModelContext, get_default_conda_env
 from mlflow.tracking.artifact_utils import _download_artifact_from_uri
+from mlflow.types import DataType, Schema
 from mlflow.utils import PYTHON_VERSION, deprecated, get_major_minor_py_version
 from mlflow.utils.file_utils import TempDir, _copy_file_or_tree
 from mlflow.utils.model_utils import _get_flavor_configuration
 from mlflow.exceptions import MlflowException
-from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE, RESOURCE_ALREADY_EXISTS
+from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE, RESOURCE_ALREADY_EXISTS, \
+    RESOURCE_DOES_NOT_EXIST
 
 FLAVOR_NAME = "python_function"
 MAIN = "loader_module"
@@ -260,7 +270,165 @@ def _load_model_env(path):
     return _get_flavor_configuration(model_path=path, flavor_name=FLAVOR_NAME).get(ENV, None)
 
 
-def load_model(model_uri, suppress_warnings=True):
+def _enforce_type(name, values: pandas.Series, t: DataType):
+    """
+    Enforce the input column type matches the declared in model input schema.
+
+    The following type conversions are allowed:
+
+    1. np.object -> string
+    2. int -> long (upcast)
+    3. float -> double (upcast)
+
+    Any other type mismatch will raise error.
+    """
+    if values.dtype == np.object and t not in (DataType.binary, DataType.string):
+        values = values.infer_objects()
+
+    if values.dtype in (t.to_pandas(), t.to_numpy()):
+        # The types are already compatible => conversion is not necessary.
+        return values
+
+    if t == DataType.binary and values.dtype.kind == t.binary.to_numpy().kind:
+        # NB: bytes in numpy have variable itemsize depending on the length of the longest
+        # element in the array (column). Since MLflow binary type is length agnostic, we ignore
+        # itemsize when matching binary columns.
+        return values
+
+    if t == DataType.string and values.dtype == np.object:
+        #  NB: strings are by default parsed and inferred as objects, but it is
+        # recommended to use StringDtype extension type if available. See
+        #
+        # `https://pandas.pydata.org/pandas-docs/stable/user_guide/text.html`
+        #
+        # for more detail.
+        try:
+            return values.astype(t.to_pandas(), errors="raise")
+        except ValueError:
+            raise MlflowException(
+                "Failed to convert column {0} from type {1} to {2}.".format(
+                  name, values.dtype, t)
+            )
+
+    numpy_type = t.to_numpy()
+    is_compatible_type = values.dtype.kind == numpy_type.kind
+    is_upcast = values.dtype.itemsize <= numpy_type.itemsize
+    if is_compatible_type and is_upcast:
+        return values.astype(numpy_type, errors="raise")
+    else:
+        # NB: conversion between incompatible types (e.g. floats -> ints or
+        # double -> float) are not allowed. While supported by pandas and numpy,
+        # these conversions alter the values significantly.
+        raise MlflowException("Incompatible input types for column {0}. "
+                              "Can not safely convert {1} to {2}.".format(name,
+                                                                          values.dtype,
+                                                                          numpy_type))
+
+
+def _enforce_schema(pdf: pandas.DataFrame, input_schema: Schema):
+    """
+    Enforce column names and types match the input schema.
+
+    For column names, we check there are no missing columns and reorder the columns to match the
+    ordering declared in schema if necessary. Any extra columns are ignored.
+
+    For column types, we make sure the types match schema or can be safely converted to match the
+    input schema.
+    """
+    if isinstance(pdf, list):
+        pdf = pandas.DataFrame(pdf)
+    if not isinstance(pdf, pandas.DataFrame):
+        message = 'Expected input to be DataFrame or list. Found: %s' % type(pdf).__name__
+        raise MlflowException(message)
+
+    if input_schema.has_column_names():
+        # make sure there are no missing columns
+        col_names = input_schema.column_names()
+        expected_names = set(col_names)
+        actual_names = set(pdf.columns)
+        missing_cols = expected_names - actual_names
+        extra_cols = actual_names - expected_names
+        # Preserve order from the original columns, since missing/extra columns are likely to
+        # be in same order.
+        missing_cols = [c for c in col_names if c in missing_cols]
+        extra_cols = [c for c in pdf.columns if c in extra_cols]
+        if missing_cols:
+            message = ("Model input is missing columns {0}."
+                       " Note that there were extra columns: {1}".format(missing_cols, extra_cols))
+            raise MlflowException(message)
+    else:
+        # The model signature does not specify column names => we can only verify column count.
+        if len(pdf.columns) < len(input_schema.columns):
+            message = ("Model input is missing input columns. The model signature declares "
+                       "{0} input columns but the provided input only has "
+                       "{1} columns. Note: the columns were not named in the signature so we can "
+                       "only verify their count.").format(len(input_schema.columns),
+                                                          len(pdf.columns))
+            raise MlflowException(message)
+        col_names = pdf.columns[:len(input_schema.columns)]
+    col_types = input_schema.column_types()
+    new_pdf = pandas.DataFrame()
+    for i, x in enumerate(col_names):
+        new_pdf[x] = _enforce_type(x, pdf[x], col_types[i])
+    return new_pdf
+
+
+PyFuncOutput = Union[pandas.DataFrame, pandas.Series, np.ndarray, list]
+
+
+class PyFuncModel(object):
+    """
+    MLflow 'python function' model.
+
+    Wrapper around model implementation and metadata. This class is not meant to be constructed
+    directly. Instead, instances of this class are constructed and returned from
+    py:func:`mlflow.pyfunc.load_model`.
+
+    ``model_impl`` can be any Python object that implements the `Pyfunc interface
+    <https://mlflow.org/docs/latest/python_api/mlflow.pyfunc.html#pyfunc-inference-api>`_, and is
+    by invoking the model's ``loader_module``.
+
+    ``model_meta`` contains model metadata loaded from the MLmodel file.
+    """
+
+    def __init__(self, model_meta: Model, model_impl: Any):
+        if not hasattr(model_impl, "predict"):
+            raise MlflowException("Model implementation is missing required predict method.")
+        if not model_meta:
+            raise MlflowException("Model is missing metadata.")
+        self._model_meta = model_meta
+        self._model_impl = model_impl
+
+    def predict(self, data: pandas.DataFrame) -> PyFuncOutput:
+        """
+        Generate model predictions.
+        :param data: Model input as pandas.DataFrame.
+        :return: Model predictions as one of pandas.DataFrame, pandas.Series, numpy.ndarray or list.
+        """
+        input_schema = self._model_meta.get_input_schema()
+        if input_schema is not None:
+            data = _enforce_schema(data, input_schema)
+        return self._model_impl.predict(data)
+
+    @property
+    def metadata(self):
+        """Model metadata."""
+        if self._model_meta is None:
+            raise MlflowException("Model is missing metadata.")
+        return self._model_meta
+
+    def __repr__(self):
+        info = {}
+        if self._model_meta is not None:
+            if self._model_meta.run_id is not None:
+                info["run_id"] = self._model_meta.run_id
+            if self._model_meta.artifact_path is not None:
+                info["artifact_path"] = self._model_meta.artifact_path
+            info["flavor"] = self._model_meta.flavors[FLAVOR_NAME]["loader_module"]
+        return yaml.safe_dump({"mlflow.pyfunc.loaded_model": info}, default_flow_style=False)
+
+
+def load_model(model_uri: str, suppress_warnings: bool = True) -> PyFuncModel:
     """
     Load a model stored in Python function format.
 
@@ -280,16 +448,23 @@ def load_model(model_uri, suppress_warnings=True):
                               loading process will be suppressed. If ``False``, these warning
                               messages will be emitted.
     """
-    local_model_path = _download_artifact_from_uri(artifact_uri=model_uri)
-    conf = _get_flavor_configuration(model_path=local_model_path, flavor_name=FLAVOR_NAME)
+    local_path = _download_artifact_from_uri(artifact_uri=model_uri)
+    model_meta = Model.load(os.path.join(local_path, "MLmodel"))
+
+    conf = model_meta.flavors.get(FLAVOR_NAME)
+    if conf is None:
+        raise MlflowException(
+            "Model does not have the \"{flavor_name}\" flavor".format(flavor_name=FLAVOR_NAME),
+            RESOURCE_DOES_NOT_EXIST)
     model_py_version = conf.get(PY_VERSION)
     if not suppress_warnings:
         _warn_potentially_incompatible_py_version_if_necessary(model_py_version=model_py_version)
     if CODE in conf and conf[CODE]:
-        code_path = os.path.join(local_model_path, conf[CODE])
+        code_path = os.path.join(local_path, conf[CODE])
         mlflow.pyfunc.utils._add_code_to_system_path(code_path=code_path)
-    data_path = os.path.join(local_model_path, conf[DATA]) if (DATA in conf) else local_model_path
-    return importlib.import_module(conf[MAIN])._load_pyfunc(data_path)
+    data_path = os.path.join(local_path, conf[DATA]) if (DATA in conf) else local_path
+    model_impl = importlib.import_module(conf[MAIN])._load_pyfunc(data_path)
+    return PyFuncModel(model_meta=model_meta, model_impl=model_impl)
 
 
 @deprecated("mlflow.pyfunc.load_model", 1.0)
@@ -340,8 +515,8 @@ def spark_udf(spark, model_uri, result_type="double"):
     """
     A Spark UDF that can be used to invoke the Python function formatted model.
 
-    Parameters passed to the UDF are forwarded to the model as a DataFrame where the names are
-    ordinals (0, 1, ...). On some versions of Spark, it is also possible to wrap the input in a
+    Parameters passed to the UDF are forwarded to the model as a DataFrame where the column names
+    are ordinals (0, 1, ...). On some versions of Spark, it is also possible to wrap the input in a
     struct. In that case, the data will be passed as a DataFrame with column names given by the
     struct definition (e.g. when invoked as my_udf(struct('x', 'y'), the model will ge the data as a
     pandas DataFrame with 2 columns 'x' and 'y').
@@ -351,8 +526,11 @@ def spark_udf(spark, model_uri, result_type="double"):
     converted to string. If the result type is not an array type, the left most column with
     matching type is returned.
 
-    >>> predict = mlflow.pyfunc.spark_udf(spark, "/my/local/model")
-    >>> df.withColumn("prediction", predict("name", "age")).show()
+    .. code-block:: python
+        :caption: Example
+
+        predict = mlflow.pyfunc.spark_udf(spark, "/my/local/model")
+        df.withColumn("prediction", predict("name", "age")).show()
 
     :param spark: A SparkSession object.
     :param model_uri: The location, in URI format, of the MLflow model with the
@@ -405,10 +583,10 @@ def spark_udf(spark, model_uri, result_type="double"):
     from mlflow.pyfunc.spark_model_cache import SparkModelCache
     from pyspark.sql.functions import pandas_udf
     from pyspark.sql.types import _parse_datatype_string
-    from pyspark.sql.types import ArrayType, DataType
+    from pyspark.sql.types import ArrayType, DataType as SparkDataType
     from pyspark.sql.types import DoubleType, IntegerType, FloatType, LongType, StringType
 
-    if not isinstance(result_type, DataType):
+    if not isinstance(result_type, SparkDataType):
         result_type = _parse_datatype_string(result_type)
 
     elem_type = result_type
@@ -430,8 +608,9 @@ def spark_udf(spark, model_uri, result_type="double"):
 
     def predict(*args):
         model = SparkModelCache.get_or_load(archive_path)
-        schema = {str(i): arg for i, arg in enumerate(args)}
+        input_schema = model.metadata.get_input_schema()
         pdf = None
+
         for x in args:
             if type(x) == pandas.DataFrame:
                 if len(args) != 1:
@@ -439,17 +618,32 @@ def spark_udf(spark, model_uri, result_type="double"):
                                     "input column, but got %d" % len(args))
                 pdf = x
         if pdf is None:
-            # Explicitly pass order of columns to avoid lexicographic ordering (i.e., 10 < 2)
-            columns = [str(i) for i, _ in enumerate(args)]
-            pdf = pandas.DataFrame(schema, columns=columns)
+            args = list(args)
+            if input_schema is None:
+                names = [str(i) for i in range(len(args))]
+            else:
+                names = input_schema.column_names()
+                if len(args) > len(names):
+                    args = args[:len(names)]
+                if len(args) < len(names):
+                    message = ("Model input is missing columns. Expected {0} input columns {1},"
+                               " but the model received only {2} unnamed input columns"
+                               " (Since the columns were passed unnamed they are expected to be in"
+                               " the order specified by the schema).".format(
+                                   len(names), names, len(args)))
+                    raise MlflowException(message)
+            pdf = pandas.DataFrame(data={names[i]: x for i, x in enumerate(args)}, columns=names)
+
         result = model.predict(pdf)
+
         if not isinstance(result, pandas.DataFrame):
             result = pandas.DataFrame(data=result)
 
-        elif type(elem_type) == IntegerType:
+        elem_type = result_type.elementType if isinstance(result_type, ArrayType) else result_type
+
+        if type(elem_type) == IntegerType:
             result = result.select_dtypes([np.byte, np.ubyte, np.short, np.ushort,
                                            np.int32]).astype(np.int32)
-
         elif type(elem_type) == LongType:
             result = result.select_dtypes([np.byte, np.ubyte, np.short, np.ushort, np.int, np.long])
 
@@ -478,7 +672,9 @@ def spark_udf(spark, model_uri, result_type="double"):
 
 
 def save_model(path, loader_module=None, data_path=None, code_path=None, conda_env=None,
-               mlflow_model=Model(), python_model=None, artifacts=None, **kwargs):
+               mlflow_model=None, python_model=None, artifacts=None,
+               signature: ModelSignature = None, input_example: ModelInputExample = None,
+               **kwargs):
     """
     save_model(path, loader_module=None, data_path=None, code_path=None, conda_env=None,\
                mlflow_model=Model(), python_model=None, artifacts=None)
@@ -556,6 +752,25 @@ def save_model(path, loader_module=None, data_path=None, code_path=None, conda_e
                       path via ``context.artifacts["my_file"]``.
 
                       If ``None``, no artifacts are added to the model.
+
+    :param signature: (Experimental) :py:class:`ModelSignature <mlflow.models.ModelSignature>`
+                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
+                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
+                      from datasets with valid model input (e.g. the training dataset with target
+                      column omitted) and valid model output (e.g. model predictions generated on
+                      the training dataset), for example:
+
+                      .. code-block:: python
+
+                        from mlflow.models.signature import infer_signature
+                        train = df.drop_column("target_label")
+                        predictions = ... # compute model predictions
+                        signature = infer_signature(train, predictions)
+    :param input_example: (Experimental) Input example provides one or several instances of valid
+                          model input. The example can be used as a hint of what data to feed the
+                          model. The given example will be converted to a Pandas DataFrame and then
+                          serialized to json using the Pandas split-oriented format. Bytes are
+                          base64-encoded.
     """
     mlflow_model = kwargs.pop('model', mlflow_model)
     if len(kwargs) > 0:
@@ -563,6 +778,7 @@ def save_model(path, loader_module=None, data_path=None, code_path=None, conda_e
     if code_path is not None:
         if not isinstance(code_path, list):
             raise TypeError('Argument code_path should be a list, not {}'.format(type(code_path)))
+
     first_argument_set = {
         "loader_module": loader_module,
         "data_path": data_path,
@@ -591,6 +807,18 @@ def save_model(path, loader_module=None, data_path=None, code_path=None, conda_e
             message=msg,
             error_code=INVALID_PARAMETER_VALUE)
 
+    if os.path.exists(path):
+        raise MlflowException(
+            message="Path '{}' already exists".format(path),
+            error_code=RESOURCE_ALREADY_EXISTS)
+    os.makedirs(path)
+    if mlflow_model is None:
+        mlflow_model = Model()
+    if signature is not None:
+        mlflow_model.signature = signature
+    if input_example is not None:
+        _save_example(mlflow_model, input_example, path)
+
     if first_argument_set_specified:
         return _save_model_with_loader_module_and_data_path(
             path=path, loader_module=loader_module, data_path=data_path,
@@ -602,7 +830,8 @@ def save_model(path, loader_module=None, data_path=None, code_path=None, conda_e
 
 
 def log_model(artifact_path, loader_module=None, data_path=None, code_path=None, conda_env=None,
-              python_model=None, artifacts=None, registered_model_name=None):
+              python_model=None, artifacts=None, registered_model_name=None,
+              signature: ModelSignature=None, input_example: ModelInputExample=None):
     """
     Log a Pyfunc model with custom inference logic and optional data dependencies as an MLflow
     artifact for the current run.
@@ -679,6 +908,25 @@ def log_model(artifact_path, loader_module=None, data_path=None, code_path=None,
                                   future release without warning. If given, create a model
                                   version under ``registered_model_name``, also creating a
                                   registered model if one with the given name does not exist.
+
+    :param signature: (Experimental) :py:class:`ModelSignature <mlflow.models.ModelSignature>`
+                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
+                      The model signature can be :py:func:`inferred <mlflow.models.infer_signature>`
+                      from datasets with valid model input (e.g. the training dataset with target
+                      column omitted) and valid model output (e.g. model predictions generated on
+                      the training dataset), for example:
+
+                      .. code-block:: python
+
+                        from mlflow.models.signature import infer_signature
+                        train = df.drop_column("target_label")
+                        predictions = ... # compute model predictions
+                        signature = infer_signature(train, predictions)
+    :param input_example: (Experimental) Input example provides one or several instances of valid
+                          model input. The example can be used as a hint of what data to feed the
+                          model. The given example will be converted to a Pandas DataFrame and then
+                          serialized to json using the Pandas split-oriented format. Bytes are
+                          base64-encoded.
     """
     return Model.log(artifact_path=artifact_path,
                      flavor=mlflow.pyfunc,
@@ -688,7 +936,9 @@ def log_model(artifact_path, loader_module=None, data_path=None, code_path=None,
                      python_model=python_model,
                      artifacts=artifacts,
                      conda_env=conda_env,
-                     registered_model_name=registered_model_name)
+                     registered_model_name=registered_model_name,
+                     signature=signature,
+                     input_example=input_example)
 
 
 def _save_model_with_loader_module_and_data_path(path, loader_module, data_path=None,
@@ -709,15 +959,9 @@ def _save_model_with_loader_module_and_data_path(path, loader_module, data_path=
                       this model should be run in.
     :return: Model configuration containing model info.
     """
-    if os.path.exists(path):
-        raise MlflowException(
-            message="Path '{}' already exists".format(path),
-            error_code=RESOURCE_ALREADY_EXISTS)
-    os.makedirs(path)
 
     code = None
     data = None
-    env = None
 
     if data_path is not None:
         model_file = _copy_file_or_tree(src=data_path, dst=path, dst_dir="data")
@@ -728,12 +972,17 @@ def _save_model_with_loader_module_and_data_path(path, loader_module, data_path=
             _copy_file_or_tree(src=code_path, dst=path, dst_dir="code")
         code = "code"
 
-    if conda_env is not None:
-        shutil.copy(src=conda_env, dst=os.path.join(path, "mlflow_env.yml"))
-        env = "mlflow_env.yml"
+    conda_env_subpath = "mlflow_env.yml"
+    if conda_env is None:
+        conda_env = get_default_conda_env()
+    elif not isinstance(conda_env, dict):
+        with open(conda_env, "r") as f:
+            conda_env = yaml.safe_load(f)
+    with open(os.path.join(path, conda_env_subpath), "w") as f:
+        yaml.safe_dump(conda_env, stream=f, default_flow_style=False)
 
     mlflow.pyfunc.add_to_model(
-        mlflow_model, loader_module=loader_module, code=code, data=data, env=env)
+        mlflow_model, loader_module=loader_module, code=code, data=data, env=conda_env_subpath)
     mlflow_model.save(os.path.join(path, 'MLmodel'))
     return mlflow_model
 
